/* gnubiff -- a mail notification program
 * Copyright (C) 2000-2003 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <sstream>
#include <cstdio>
#include <string>
#include <glib.h>

#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "Mailbox.h"


// ===================================================================
// = Static attributes ===============================================
// ===================================================================


// ===================================================================
// = Callbacks "C" binding ===========================================
// ===================================================================


// ===================================================================
// = Mailbox ========================================================
// ===================================================================
Mailbox::Mailbox (Biff *owner) {
  _owner = owner;
  _unreads.clear();
  _status = MAILBOX_UNITIALIZED;
  _sd = SD_CLOSE;

#ifdef HAVE_LIBSSL
  // SSL initialization
  _ssl = 0;  
  if (_context == 0) {
    SSL_library_init ();
    SSL_load_error_strings();
    _context = SSL_CTX_new (SSLv23_client_method());
    if (!_context) 
      g_warning ("Problem initializing SSL library (couldn't create context)");
  }
#endif
}

#ifdef HAVE_LIBSSL
SSL_CTX *Mailbox::_context = 0;
#endif

// ===================================================================
// = ~Mailbox ========================================================
// ===================================================================
Mailbox::~Mailbox (void) {
}

// ===================================================================
// = parse ===========================================================
/**
   Parse a mail and fill a header structure accordingly
   @param mail to parse
   @param status of the mail if known
 */
// ===================================================================
void Mailbox::parse (vector<string> &mail, int status) {
  header h;
  h.status = status;

  for (guint i=0; i<mail.size(); i++) {
    gchar *buffer = g_ascii_strdown (mail[i].c_str(), -1);
    string line = buffer;
    g_free (buffer);

    // Sender
    // There should a whitespace or a tab after "From:", so we look
    // for "From:" and get string beginning at 6
    if ((mail[i].find ("From:") == 0) && h.sender.empty())
      h.sender = mail[i].substr (6);

    // Subject
    // There should a whitespace or a tab after "Subject:", so we look
    // for "Subject:" and get string beginning at 9
    else if ((mail[i].find ("Subject:") == 0) && h.subject.empty())
      h.subject = mail[i].substr (9);

    // Date
    // There should a whitespace or a tab after "Date:", so we look
    // for "Date:" and get string beginning at 6
    else if ((mail[i].find ("Date:") == 0) && h.date.empty())
      h.date = mail[i].substr (6);

    // Charset
    // FIXME: Currently, this code does not handle the case where a
    // charset is coded over two or more lines. Question: is is a big
    // problem ? (is it allowed anyway ?)
    else if ((line.find ("charset=") != string::npos) && h.charset.empty()) {
      string charset = line.substr (int(line.find ("charset="))+8);
      // First we remove any leading '"'
      if (charset[0] == '\"')
	charset = charset.substr(1);

      // Then we wait for the end of charset
      for (guint j=0; j<charset.size(); j++) {
	if ((charset[j+1] == ';') || (charset[j+1] == '\"') ||
	    (charset[j+1] == '\n') ||  (charset[j+1] == '\t') ||
	    (charset[j+1] == ' ') ||  (charset[j+1] == '\0')) {
	  h.charset = charset.substr (0, j+1);
	  break;
	}
      }
    }
    // Status
    else if ((mail[i].find ("Status: R") == 0) && h.status == -1)
      h.status = MAIL_READ;
    else if ((mail[i].find ("X-Mozilla-Status: 0001") == 0) && h.status == -1)
      h.status = MAIL_READ;

    else if ((mail[i].empty()) && h.body.empty()) {
      guint j = 0;
      do {
	h.body += mail[i++] + string("\n");
	j++;
      } while ((j<10) && (i < mail.size()));
      if (j == 10)
	h.body += string("...");
    }
  }

  // Store mail depending on status
  if ((h.status == MAIL_UNREAD) || (h.status == -1))
    // Pines trick
    if (h.subject.find("DON'T DELETE THIS MESSAGE -- FOLDER INTERNAL DATA") == string::npos)
      _unreads.push_back(h);
}



// ===================================================================
// = socket_open =====================================================
/**
   Open a socket
   @param hostname is the host to contact to
   @param port is the port to connect to
 */
// ===================================================================
void Mailbox::socket_open (string &hostname, unsigned short int port) {
  struct sockaddr_in sin;
  struct hostent *host;
  struct in_addr address;

  // Default status before trying to connect
  _socket_status = SOCKET_STATUS_ERROR;

  // Create an endpoint for communication
  if ((_sd = socket (AF_INET, SOCK_STREAM, IPPROTO_IP)) == -1){
    _sd = SD_CLOSE;
    return;
  }

  // Setting socket info for connection
  memset ((char *)&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons (port);

  // First, try to get the address by standard notation (e.g. 127.0.0.1)
  if (inet_aton (hostname.c_str(), &address) == 0) {
    // If it does not work, get the address by name (e.g. localhost.localdomain)
    if ((host = gethostbyname (hostname.c_str())) == 0)
      return;

    // This way of filling sin_addr field avoid 'incompatible types in assignment' problems
    memcpy ((void *) &sin.sin_addr, *(host->h_addr_list), host->h_length);
  }
  else
    sin.sin_addr = address;
//    memcpy ((void *) &sin.sin_addr, (void *)&address, sizeof (address));

  // Initiate the connection on the socket
  if ((connect (_sd, (struct sockaddr *) &sin, sizeof (struct sockaddr_in))) == -1) {
    close (_sd);
    _sd=SD_CLOSE;
    return;
  }

#ifdef HAVE_LIBSSL
  if (_owner->_ssl) {
    if (_owner->_certificate.size() > 0){
      if (SSL_CTX_load_verify_locations(_context, _owner->_certificate.c_str(),	NULL) == 0) {
        g_warning("Failed to load server certificate (%s)", _owner->_certificate.c_str());
        close (_sd);
        _sd = SD_CLOSE;
        return;
      }
      SSL_CTX_set_verify (_context, SSL_VERIFY_PEER, NULL);
    }
    else {
      SSL_CTX_set_verify (_context, SSL_VERIFY_NONE, NULL);
    }
    
    _ssl = SSL_new (_context);
    if (!_ssl){
      g_warning("Failed to connect to server");
      close (_sd);
      _sd = SD_CLOSE;
      return;
    }
    if (SSL_set_fd (_ssl, _sd) == 0) {
      g_warning ("Failed to connect to server");
      close (_sd);
      _sd = SD_CLOSE;
      return;
    }
    if (SSL_connect (_ssl) != 1){
      g_warning("Failed to connect to server");
      SSL_free (_ssl);
      _ssl = NULL;
      close (_sd);
      _sd = SD_CLOSE;
      return;
    }
  }
#endif

  _socket_status = SOCKET_STATUS_OK;
  return;
};


// ===================================================================
// = socket_close ====================================================
/**
   Close the current socket (_sd)
 */
// ===================================================================
void Mailbox::socket_close (void) {
  string line;
  
  if(_sd != SD_CLOSE) {
    fcntl (_sd, F_SETFL, O_NONBLOCK);
    do {
      socket_read (line);
    } while (!line.empty());
  }
  
#ifdef HAVE_LIBSSL
  if (_owner->_ssl && _ssl) {
    if (_sd != SD_CLOSE)
      SSL_shutdown (_ssl);
    SSL_free (_ssl);
    _ssl = 0;
  }
#endif
  
  if(_sd != SD_CLOSE)
    close (_sd);
  _sd = SD_CLOSE;
}


// ===================================================================
// = socket_write ====================================================
/**
   Write through a socket
   @param line is the line to write
 */
// ===================================================================
void Mailbox::socket_write (string line) {
  _socket_status = SOCKET_STATUS_ERROR;

//  cout << "WRITE: " << line << endl;

#ifdef HAVE_LIBSSL
  if (_owner->_ssl) {
    if (SSL_write (_ssl, line.c_str(), line.size()) <= 0)
      return;
    _socket_status = SOCKET_STATUS_OK;
    return;
  }
#endif
  if (write (_sd, line.c_str(), line.size()) <= 0)
    return;
  _socket_status = SOCKET_STATUS_OK;
}

// ===================================================================
// = socket_read =====================================================
/**
   Read from a socket
   @param line is the read line
 */
// ===================================================================
void Mailbox::socket_read (string &line) {
  char buffer;
  int status;
  line = "";
  _socket_status = SOCKET_STATUS_ERROR;


#ifdef HAVE_LIBSSL
  if (_owner->_ssl) {
    while (((status = SSL_read (_ssl, &buffer, 1)) > 0) && (buffer != '\n'))
      line += buffer;
    if (status > 0) 
      _socket_status = SOCKET_STATUS_OK;
//    cout << "READ: " << line << endl;
    return;
  }
#endif
  while (((status = read (_sd, &buffer, 1)) > 0) && (buffer != '\n'))
    line += buffer;
  if (status > 0) 
    _socket_status = SOCKET_STATUS_OK;
//  cout << "READ: " << line << endl;
}
